import time
import asyncio


class Proxy:
    def __init__(self):
        self.gateway_write = None
        self.gateway_write_eof = None
        self.gateway_close = None
        self.target_write = None
        self.target_write_eof = None
        self.target_get_returncode = None
        self.history = []
        self.exited = asyncio.Future()

    def attach_channel(self, channel, target_write, target_write_eof, target_get_returncode=None):
        attached_channel = channel
        attached_target_write = target_write
        attached_target_write_eof = target_write_eof
        original_target_write = self.target_write
        original_target_write_eof = self.target_write_eof
        def target_write(data, channel):
            if original_target_write:
                original_target_write(data, channel)
            if channel == attached_channel:
                attached_target_write(data)
        def target_write_eof(channel):
            if original_target_write_eof:
                original_target_write_eof(channel)
            if channel == attached_channel:
                attached_target_write_eof()
        self.target_write = target_write
        self.target_write_eof = target_write_eof
        if target_get_returncode:
            self.target_get_returncode = target_get_returncode

    def on_recv(self, data, channel=None):
        self.history.append({
            'direction': 'output',
            'data': data,
            'channel': channel,
            'time': time.time()
        })
        if data and self.gateway_write:
            self.gateway_write(data, channel)
        return data

    def on_send(self, data, channel=None):
        self.history.append({
            'direction': 'input',
            'data': data,
            'channel': channel,
            'time': time.time()
        })
        if data and self.target_write:
            self.target_write(data, channel)
        elif not data and self.target_write_eof:
            self.target_write_eof(channel)
        return data

    def on_exit(self):
        if self.exited.done():
            return
        history = [
            {
                k: v if type(v) is not bytes else v.decode('latin')
                for k, v in e.items()
            }
            for e in self.history
        ]
        data = {
            'interaction': history,
            'return_code': self.target_get_returncode() if self.target_get_returncode else None
        }
        if self.gateway_close:
            self.gateway_close()
        self.exited.set_result(data)
        return data



import sys
import json
import pathlib
import asyncio

import aiohttp
import aiohttp.web

import pwnstar.tubes


async def run_server(create_target, create_proxy, *, host, port):
    loop = asyncio.get_running_loop()

    def protocol_factory():
        proxy = create_proxy()
        return pwnstar.tubes.GatewayProtocol(create_target, proxy)

    server = await loop.create_server(protocol_factory, host=host, port=port)

    async with server:
        await server.serve_forever()


async def run_webserver(create_target, create_proxy, *, host, port, channels, start=True):
    loop = asyncio.get_running_loop()
    static_dir = pathlib.Path(__file__).parent / 'ws_static'

    async def index_handler(request):
        return aiohttp.web.FileResponse(static_dir / 'index.html')

    async def info_handler(request):
        return aiohttp.web.json_response({
            'channels': channels
        })

    async def ws_handler(request):
        ws = aiohttp.web.WebSocketResponse()
        await ws.prepare(request)

        proxy = create_proxy()

        async def gateway_write_json(data, channel=None):
            await ws.send_bytes(json.dumps({
                'data': data.decode('latin'),
                'channel': channel
            }).encode())

        proxy.gateway_write = lambda data, channel: loop.create_task(gateway_write_json(data, channel))
        proxy.gateway_write_eof = lambda channel: loop.create_task(ws.close())
        proxy.gateway_close = lambda: loop.create_task(ws.close())

        target = await create_target(proxy)

        await ws.send_bytes(json.dumps({
            'status': 'ready'
        }).encode())

        async for msg in ws:
            json_data = json.loads(msg.data)
            data = json_data['data'].encode('latin')
            channel = json_data['channel']
            if data:
                proxy.on_send(data, channel)
            else:
                proxy.on_send(b'', channel)

    app = aiohttp.web.Application()
    app.add_routes([aiohttp.web.get('/', index_handler),
                    aiohttp.web.get('/info', info_handler),
                    aiohttp.web.get('/ws', ws_handler),
                    aiohttp.web.static('/',  static_dir, show_index=True)])

    if start:
        await aiohttp.web._run_app(app, host=host, port=port)
    else:
        return app


async def run_local(create_target, create_proxy):
    loop = asyncio.get_running_loop()

    proxy = create_proxy()

    gateway = pwnstar.tubes.GatewayProtocol(create_target, proxy)

    await loop.connect_read_pipe(lambda: gateway, sys.stdin)
    await loop.connect_write_pipe(lambda: gateway, sys.stdout)

    data = await proxy.exited
    return data['return_code']



import os
import time
import logging
import asyncio

from .utils import log


logger = logging.getLogger('pwnstar.tubes.process')


class ProcessProtocol(asyncio.SubprocessProtocol):
    def __init__(self, proxy):
        self.proxy = proxy

    @log(logger)
    def pipe_data_received(self, fd, data):
        self.proxy.on_recv(data, fd)

    @log(logger)
    def pipe_connection_lost(self, fd, exc):
        self.proxy.on_recv(b'', fd)

    @log(logger)
    def process_exited(self):
        self.proxy.on_exit()



import os
import pty
import asyncio

import pwnstar.tubes


async def create_process_target(proxy, *, proc_args):
    loop = asyncio.get_running_loop()

    target_transport, target_protocol = await loop.subprocess_exec(
        lambda: pwnstar.tubes.ProcessProtocol(proxy),
        *proc_args,
        close_fds=False,
        env=os.environ)

    stdin = target_transport.get_pipe_transport(0)
    proxy.attach_channel(0,
                         stdin.write,
                         stdin.write_eof,
                         target_transport.get_returncode)


async def create_tty_process_target(proxy, *, proc_args):
    loop = asyncio.get_running_loop()

    master, slave = pty.openpty()

    target_transport, target_protocol = await loop.subprocess_exec(
        lambda: pwnstar.tubes.ProcessProtocol(proxy),
        *proc_args,
        close_fds=False,
        stdin=slave,
        stdout=slave,
        stderr=slave,
        start_new_session=True,
        env=os.environ)

    target_transport._proc.stdin = open(master, 'wb', -1)
    target_transport._proc.stdout = open(master, 'rb', -1)
    target_transport._proc.stderr = open(master, 'rb', -1)
    await target_transport._connect_pipes(None)

    def target_write(data):
        target_transport._proc.stdin.write(data)
        target_transport._proc.stdin.flush()

    proxy.attach_channel(0,
                         target_write,
                         lambda: target_write(b'\x04'),
                         target_transport.get_returncode)


async def create_remote_target(proxy, *, host=None, port=None, sock=None, channel=None):
    loop = asyncio.get_running_loop()

    if not channel:
        channel = f'{host}:{port}'

    target_transport, target_protocol = await loop.create_connection(
        lambda: pwnstar.tubes.RemoteProtocol(proxy, channel=channel),
        host=host,
        port=port,
        sock=sock)

    proxy.attach_channel(channel,
                         target_transport.write,
                         target_transport.write_eof)



import logging
import asyncio

from .utils import log


logger = logging.getLogger('pwnstar.tubes.gateway')


class GatewayProtocol(asyncio.Protocol):
    def __init__(self, create_target, proxy, channel=None):
        self.create_target = create_target
        self.proxy = proxy
        self.channel = channel

    def connection_made(self, transport):
        if isinstance(transport, asyncio.WriteTransport):
            self.proxy.gateway_write = lambda data, fd: transport.write(data)
            self.proxy.gateway_write_eof = lambda fd: transport.write_eof()
            self.proxy.gateway_close = transport.close
            asyncio.get_running_loop().create_task(self.create_target(self.proxy))

    @log(logger)
    def data_received(self, data):
        self.proxy.on_send(data, self.channel)

    @log(logger)
    def eof_received(self):
        self.proxy.on_send(b'', self.channel)
        return True  # keep connection open



import functools
import inspect


def log(logger):
    def wrapper(func):
        signature = inspect.signature(func)
        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            bound = signature.bind(*args, **kwargs)
            args_info = ', '.join([repr(arg) for arg in bound.args] +
                                  [f'{repr(key)}={repr(val)}' for key, val in bound.kwargs.items()])
            logger.debug(f"{func.__qualname__}({args_info})")
            return func(*args, **kwargs)
        return wrapped
    return wrapper



import logging
import asyncio

from .utils import log


logger = logging.getLogger("pwnstar.tubes.remote")


class RemoteProtocol(asyncio.Protocol):
    def __init__(self, proxy, channel=None):
        self.proxy = proxy
        self.channel = channel

    @log(logger)
    def data_received(self, data):
        self.proxy.on_recv(data, self.channel)

    @log(logger)
    def eof_received(self):
        self.proxy.on_recv(b'', self.channel)
        self.proxy.on_exit()



import os
import subprocess


try:
    from setuptools import setup
    from setuptools import find_packages
    packages = find_packages()
except ImportError:
    from distutils.core import setup
    packages = [x.strip('./').replace('/', '.') for x in os.popen('find -name "__init__.py" | xargs -n1 dirname').read().strip().split('\n')]

setup(
    name='pwnstar',
    python_requires='>=3.7',
    version='0.3.5',
    packages=packages,
    install_requires=[
        'aiohttp'
    ],
    dependency_links=[
    ],
    package_data={
        'pwnstar': ['ws_static/*', 'ws_static/*/*']
    },
    entry_points = {
        'console_scripts': [
            'pwnstar = pwnstar.pwnstar:main',
        ],
    }
)
